/*
 * defRead.c --      
 *
 * This module incorporates the LEF/DEF format for standard-cell place and
 * route.
 *
 * Version 0.1 (September 26, 2003):  DEF input of designs.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /ufs/repository/magic/lef/defRead.c,v 1.0 2003/05/01 10:37:00 tim Exp $";            
#endif  /* not lint */

#include <stdio.h>
#include <stdlib.h>
#ifdef SYSV
#include <string.h>
#endif
#include <errno.h>
#include <stdarg.h>

#include "magic/tclmagic.h"
#include "misc/magic.h"
#include "utils/geometry.h"
#include "tiles/tile.h"
#include "utils/hash.h"
#include "database/database.h"
#include "windows/windows.h"
#include "dbwind/dbwind.h"
#include "utils/malloc.h"
#include "graphics/graphics.h"
#include "main/main.h"
#include "cif/cif.h"
#include "lef/lefInt.h"

/*
 *------------------------------------------------------------
 *
 * DefAddRoutes --
 *
 *	Parse a network route statement from the DEF file,
 *	and add it to the linked list representing the route.
 *
 * Results:
 *	Returns the position in the intput line ("lineptr")
 *	after parsing a single routing statement ("+" keyword).
 *
 * Side Effects:
 *	Adds information to the layout database.
 *
 *
 *------------------------------------------------------------
 */

#define DEFAULT_WIDTH 3

char *
DefAddRoutes(rootDef, lineptr, oscale, special)
    CellDef *rootDef;		/* Cell to paint */
    char *lineptr;		/* Input stream */
    float oscale;		/* Scale factor between LEF and magic units */
    bool special;		/* True if this section is SPECIALNETS */
{
    char *tptr = lineptr;
    char token[256];
    linkedRect *routeList, *newRoute = NULL, *routeTop = NULL;
    Point refp;			/* reference point */
    bool valid = FALSE;		/* is there a valid reference point? */
    bool initial = TRUE;
    Rect locarea;
    float x, y, w;
    int curWidth, w1, w2;
    TileType newLayer;

    while (*tptr != '\0')
    {
	/* Just in case we're not at the beginning of the next token */
        while(isspace(*tptr))
	   if (*++tptr == '\0') break;

	/* Get next point, token "NEW", or via name */
	if (initial || !strncmp(tptr, "NEW", 3) || !strncmp(tptr, "new", 3))
	{
	    /* initial pass is like a NEW record, but has no NEW keyword */
	    initial = FALSE;

	    /* invalidate reference point */
	    valid = FALSE;

	    tptr = LefSkipToken(tptr);
	    sscanf(tptr, "%s", token);
	    tptr = LefSkipToken(tptr);

	    /* Note that, like in lefRead.c, this measure is ad-hoc */

	    newLayer = DBTechNameType(LefLower(token));
	    if (newLayer < 0)
	    {
		LefError("Unknown layer type \"%s\" for NEW route\n", token); 
		continue;
	    }

	    if (special)
	    {
		/* SPECIALNETS has the additional width */
		if (sscanf(tptr, "%f", &w) != 1)
		{
		    LefError("Bad width in special net\n");
		    continue;
		}
		tptr = LefSkipToken(tptr);
		curWidth = (w == 0) ? DEFAULT_WIDTH : (int)(w / oscale);
	    }
	    else
		curWidth = DEFAULT_WIDTH;
	}
	else if (*tptr != '(')	/* via name */
	{
	    sscanf(tptr, "%255s", token);
	    tptr = LefSkipToken(tptr);
	    if (valid == FALSE)
	    {
		LefError("Route has via name \"%s\" but no points!\n", token);
		continue;
	    }
	    newLayer = DBTechNameType(LefLower(token));
	    if (newLayer >= 0)
	    {
		newRoute = (linkedRect *)mallocMagic(sizeof(linkedRect));

		/* The area to paint is supposed to be derived from the	*/
		/* via definitions, not the width (to be implemented)!	*/

		w1 = curWidth >> 1;
		w2 = curWidth - w1;
		newRoute->area.r_xbot = refp.p_x - w1;
		newRoute->area.r_ybot = refp.p_y - w1;
		newRoute->area.r_xtop = refp.p_x + w2;
		newRoute->area.r_ytop = refp.p_y + w2;
	    }
	    else
		LefError("Via name \"%s\" unknown in route.\n", token);
	}
	else
	{
	    /* Record current reference point */
	    locarea.r_xbot = refp.p_x;
	    locarea.r_ybot = refp.p_y;

	    /* Read an (X Y) point */
	    tptr++;
	    while(isspace(*tptr)) tptr++;
	    if (*tptr == '*')
	    {
		if (valid == FALSE)
		{
		    LefError("No reference point for \"*\" wildcard\n"); 
		    goto endCoord;
		}
		tptr++;
	    }
	    else if (sscanf(tptr, "%f", &x) == 1)
	    {
		refp.p_x = (int)(x / oscale);
	        while(!isspace(*tptr)) tptr++;
	    }
	    else
	    {
		LefError("Cannot parse X coordinate.\n"); 
		goto endCoord;
	    }
	    while(isspace(*tptr)) tptr++;
	    if (*tptr == '*')
	    {
		if (valid == FALSE)
		{
		    LefError("No reference point for \"*\" wildcard\n"); 
		    freeMagic(newRoute);
		    newRoute = NULL;
		    goto endCoord;
		}
		tptr++;
	    }
	    else if (sscanf(tptr, "%f", &y) == 1)
	    {
		refp.p_y = (int)(y / oscale);
	        while(!isspace(*tptr) && (*tptr != ')')) tptr++;
	    }
	    else
	    {
		LefError("Cannot parse Y coordinate.\n"); 
		goto endCoord;
	    }

	    while(isspace(*tptr)) tptr++;
	    if (*tptr != ')')
	    {
		/* non-default route extension */
		/* (to be done) */
	    }

	    /* Indicate that we have a valid reference point */

	    if (valid == FALSE)
	    {
		valid = TRUE;
	    }
	    else if ((locarea.r_xbot != refp.p_x) && (locarea.r_ybot != refp.p_y))
	    {
		/* Skip over nonmanhattan segments, reset the reference	*/
		/* point, and output a warning.				*/

		LefError("Can't deal with nonmanhattan geometry in route.\n");
		locarea.r_xbot = refp.p_x;
		locarea.r_ybot = refp.p_y;
	    }
	    else
	    {
		newRoute = (linkedRect *)mallocMagic(sizeof(linkedRect));

		/* Route coordinates become the centerline of the	*/
		/* segment (to the nearest integer; may be 1/2 off. . .	*/
		/* roundoff error should be fixed!			*/

		w1 = curWidth >> 1;
		w2 = curWidth - w1;

		locarea.r_xtop = refp.p_x;
		locarea.r_ytop = refp.p_y;

		GeoCanonicalRect(&locarea, &newRoute->area);

		newRoute->area.r_xbot -= w1;
		newRoute->area.r_ybot -= w1;
		newRoute->area.r_xtop += w2;
		newRoute->area.r_ytop += w2;
	    }

endCoord:
	    /* Find the closing parenthesis for the coordinate pair */
	    while(*tptr != ')')
	    {
		if (*tptr == '\0')
		{
		    LefError("Line ended without end of coordinate.\n");
		    return tptr;
		}
		tptr++;
	    }
	    tptr = LefSkipToken(tptr);
	}

	/* Link in the new route segment */
	if (newRoute)
	{
	    newRoute->type = newLayer;
	    newRoute->rect_next = NULL;

	    if (routeTop)
		routeList->rect_next = newRoute;
	    else
		routeTop = newRoute;

	    routeList = newRoute;
	    newRoute = NULL;
	}
    }

    /* Process each segment and paint into the layout */

    while (routeTop != NULL)
    {
	/* paint */
	DBPaint(rootDef, &routeTop->area, routeTop->type);

	/* advance to next point and free record (1-delayed) */
	freeMagic((char *)routeTop);
	routeTop = routeTop->rect_next;
    }

    return tptr;
}

/*
 *------------------------------------------------------------
 *
 * DefReadNets --
 *
 *	Read a NETS or SPECIALNETS section from a DEF file.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Networks are created, and geometry may be
 *	painted into the database top-level cell.
 *
 *------------------------------------------------------------
 */

enum def_net_keys {DEF_NET_START = 0, DEF_NET_END};
enum def_netprop_keys {
	DEF_NETPROP_USE = 0, DEF_NETPROP_ROUTED, DEF_NETPROP_FIXED,
	DEF_NETPROP_COVER, DEF_NETPROP_SOURCE, DEF_NETPROP_WEIGHT,
	DEF_NETPROP_PROPERTY};

void
DefReadNets(f, rootDef, sname, oscale, special, total)
    FILE *f;
    CellDef *rootDef;
    char *sname;
    float oscale;
    bool special;		/* True if this section is SPECIALNETS */
    int total;
{
    char token[256], *tptr;
    char *lineptr;
    int keyword, subkey;
    int processed = 0;

    static char *net_keys[] = {
	"-",
	"END",
	NULL
    };

    static char *net_property_keys[] = {
	"USE",
	"ROUTED",
	"FIXED",
	"COVER",
	"SOURCE",
	"WEIGHT",
	"PROPERTY",
	NULL
    };

    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;   /* Blank line or null statement */

	keyword = Lookup(token, net_keys);
	if (keyword < 0)
	{
	    LefError("Unknown keyword \"%s\" in NET "
			"definition; ignoring.\n", token);
	    continue;
	}
	tptr = lineptr;

	switch (keyword)
	{
	    case DEF_NET_START:
		lineptr = LefComplete(f);	/* read to semicolon */

		/* Skip the "-" keyword */
		tptr = LefSkipToken(lineptr);

		/* Update the record of the number of nets processed	*/
		/* and spit out a message for every 5% finished.	*/

		LefEstimate(processed++, total,
			(special) ? "special nets" : "nets");

		/* Presently, we ignore net names completely.	*/
		/* Also, all connections are ignored, and we go	*/
		/* go directly to the first property ("+") key	*/

		/* Process all properties */
		while ((tptr = strchr(tptr, '+')) != NULL)
		{
		    /* Skip the "+" keyword */
		    tptr = LefSkipToken(tptr);
		    if (sscanf(tptr, "%s", token) != 1)
		    {
			LefError("Bad component property\n");
			break;
		    }

		    subkey = Lookup(token, net_property_keys);
		    if (subkey < 0)
		    {
			LefError("Unknown net property \"%s\" in "
				"NET definition; ignoring.\n", token);
			continue;
		    }
		    switch (subkey)
		    {
			case DEF_NETPROP_USE:
			    /* Presently, we ignore this */
			    break;
			case DEF_NETPROP_ROUTED:
			case DEF_NETPROP_FIXED:
			case DEF_NETPROP_COVER:
			    tptr = DefAddRoutes(rootDef, tptr, oscale, special);
			    break;
		    }
		}
		break;

	    case DEF_NET_END:
		if (!LefParseEndStatement(lineptr, sname))
		{
		    LefError("Net END statement missing.\n");
		    keyword = -1;
		}
		break;
	}
	if (keyword == DEF_NET_END) break;
    }

    if (processed == total)
	TxPrintf("  Processed %d%s nets total.\n", processed,
		(special) ? " special" : "");
    else
	LefError("Warning:  Number of subcells read (%d) does not match "
		"the number declared (%d).\n", processed, total);
}

/*
 *------------------------------------------------------------
 *
 * DefReadUseLocation --
 *
 *	Read location and orientation of a cell use
 *	Syntax: ( X Y ) O
 *
 *------------------------------------------------------------
 */
enum def_orient {DEF_NORTH, DEF_SOUTH, DEF_EAST, DEF_WEST,
	DEF_FLIPPED_NORTH, DEF_FLIPPED_SOUTH, DEF_FLIPPED_EAST,
	DEF_FLIPPED_WEST};

void
DefReadLocation(use, lineptr, oscale)
    CellUse *use;
    char *lineptr;
    float oscale;
{
    Rect *r;
    int keyword;
    char token[5];
    float x, y;
    Transform t;

    static char *orientations[] = {
	"N", "S", "E", "W", "FN", "FS", "FE", "FW"
    };

    if (sscanf(lineptr, "( %f %f ) %4s", &x, &y, token) != 3)
    {
	LefError("Badly formed macro orientation and location.\n");
	return;
    }

    keyword = Lookup(token, orientations);
    if (keyword < 0)
    {
	LefError("Unknown macro orientation \"%s\".\n", token);
	return;
    }

    /* The standard transformations are all defined to rotate	*/
    /* or flip about the origin.  However, DEF defines them	*/
    /* around the center such that the lower left corner is	*/
    /* unchanged after the transformation.  Case conditions	*/
    /* restore the lower-left corner position.			*/

    r = &use->cu_def->cd_bbox;
    switch (keyword)
    {
	case DEF_NORTH:
	    t = GeoIdentityTransform;
	    t.t_c = -(r->r_xbot);
	    t.t_f = -(r->r_ybot);
	    break;
	case DEF_SOUTH:
	    t = Geo180Transform;
	    t.t_c = r->r_xtop;
	    t.t_f = r->r_ytop;
	    break;
	case DEF_EAST:
	    t = Geo90Transform;
	    t.t_c = -(r->r_ybot);
	    t.t_f = r->r_xtop;
	    break;
	case DEF_WEST:
	    t = Geo270Transform;
	    t.t_c = r->r_ytop;
	    t.t_f = -(r->r_xbot);
	    break;
	case DEF_FLIPPED_NORTH:
	    t = GeoSidewaysTransform;
	    t.t_c = r->r_xtop;
	    t.t_f = -(r->r_ybot);
	    break;
	case DEF_FLIPPED_SOUTH:
	    t = GeoUpsideDownTransform;
	    t.t_c = -(r->r_xbot);
	    t.t_f = r->r_ytop;
	    break;
	case DEF_FLIPPED_EAST:
	    t = GeoRef135Transform;
	    t.t_c = r->r_ytop;
	    t.t_f = r->r_xtop;
	    break;
	case DEF_FLIPPED_WEST:
	    t = GeoRef45Transform;
	    t.t_c = -(r->r_ybot);
	    t.t_f = -(r->r_xbot);
	    break;
    }

    /* Translate --- see GeoTranslateTrans() in utils/geometry.c */
    /* t_c -> x translation, t_f -> y translation */

    t.t_c += (int)(x / oscale);
    t.t_f += (int)(y / oscale);

    DBSetTrans(use, &t);
}

/*
 *------------------------------------------------------------
 *
 * DefReadComponents --
 *
 *	Read a COMPONENTS section from a DEF file.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Cell definitions and uses are created and added to
 *	the database.
 *
 *------------------------------------------------------------
 */

enum def_comp_keys {DEF_COMP_START = 0, DEF_COMP_END};
enum def_prop_keys {
	DEF_PROP_FIXED = 0, DEF_PROP_COVER,
	DEF_PROP_PLACED, DEF_PROP_UNPLACED,
	DEF_PROP_SOURCE, DEF_PROP_WEIGHT, DEF_PROP_FOREIGN,
	DEF_PROP_REGION, DEF_PROP_GENERATE, DEF_PROP_PROPERTY,
	DEF_PROP_EEQMASTER};

void
DefReadComponents(f, rootDef, sname, oscale, total)
    FILE *f;
    CellDef *rootDef;
    char *sname;
    float oscale;
    int total;
{
    CellDef *defMacro;
    CellUse *defUse;
    char token[256], *tptr;
    char *lineptr;
    char defname[256];
    int keyword, subkey;
    int processed = 0;

    static char *component_keys[] = {
	"-",
	"END",
	NULL
    };

    static char *property_keys[] = {
	"FIXED",
	"COVER",
	"PLACED",
	"UNPLACED",
	"SOURCE",
	"WEIGHT",
	"FOREIGN",
	"REGION",
	"GENERATE",
	"PROPERTY",
	"EEQMASTER",
	NULL
    };

    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;   /* Blank line or null statement */
	keyword = Lookup(token, component_keys);

	if (keyword < 0)
	{
	    LefError("Unknown keyword \"%s\" in COMPONENT "
			"definition; ignoring.\n", token);
	    continue;
	}
	switch (keyword)
	{
	    case DEF_COMP_START:
		lineptr = LefComplete(f);	/* read to semicolon */

		/* Skip the "-" keyword */
		tptr = LefSkipToken(lineptr);

		/* Update the record of the number of components	*/
		/* processed and spit out a message for every 5% done.	*/
 
		LefEstimate(processed++, total, "subcell instances");

		/* Get use and macro names */
		if (sscanf(tptr, "%s %s", token, defname) != 2)
		{
		    LefError("Bad component statement:  Need use and macro names\n");
		    break;
		}

		/* Find the corresponding macro definition */
		defMacro = DBCellLookDef(defname);
		if (defMacro == (CellDef *)NULL)
		{
		    LefError("Cell %s is not defined.  Maybe you have not "
				"read the corresponding LEF file?\n",
				defname);
		    break;
		}

		/* Create a use for this celldef in the edit cell */
		/* Don't process properties for cells we could not find */
		if ((defUse = DBCellNewUse(defMacro, token)) == NULL)
		    break;

		/* Now do a search through the line for "+" entries	*/
		/* And process each.					*/

		while ((tptr = strchr(tptr, '+')) != NULL)
		{
		    /* Skip the "+" keyword */
		    tptr = LefSkipToken(tptr);
		    if (sscanf(tptr, "%s", token) != 1)
		    {
			LefError("Bad component property\n");
			break;
		    }

		    subkey = Lookup(token, property_keys);
		    if (subkey < 0)
		    {
			LefError("Unknown component property \"%s\" in "
				"COMPONENT definition; ignoring.\n", token);
			continue;
		    }
		    switch (subkey)
		    {
			case DEF_PROP_PLACED:
			case DEF_PROP_UNPLACED:
			case DEF_PROP_FIXED:
			case DEF_PROP_COVER:
			    tptr = LefSkipToken(tptr);
			    DefReadLocation(defUse, tptr, oscale);
			    break;
			case DEF_PROP_SOURCE:
			case DEF_PROP_WEIGHT:
			case DEF_PROP_FOREIGN:
			case DEF_PROP_REGION:
			case DEF_PROP_GENERATE:
			case DEF_PROP_PROPERTY:
			case DEF_PROP_EEQMASTER:
			    break;
		    }
		}

		/* Place the cell */
		if (defUse != NULL)
		    DBPlaceCell(defUse, rootDef);
		break;

	    case DEF_COMP_END:
		if (!LefParseEndStatement(lineptr, sname))
		{
		    LefError("Component END statement missing.\n");
		    keyword = -1;
		}

		/* Finish final call by placing the cell use */
		if (defUse != NULL)
		    DBPlaceCell(defUse, rootDef);

		break;
	}
	if (keyword == DEF_COMP_END) break;
    }

    if (processed == total)
	TxPrintf("  Processed %d subcell instances total.\n", processed);
    else
	LefError("Warning:  Number of subcells read (%d) does not match "
		"the number declared (%d).\n", processed, total);
}

/*
 *------------------------------------------------------------
 *
 * DefRead --
 *
 *	Read a .def file into a magic layout.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Cell definitions and uses are created and added to
 *	the database.
 *
 *------------------------------------------------------------
 */

/* Enumeration of sections defined in DEF files */

enum def_sections {DEF_VERSION = 0, DEF_NAMESCASESENSITIVE,
	DEF_UNITS, DEF_DESIGN, DEF_REGIONS, DEF_ROW, DEF_TRACKS,
	DEF_GCELLGRID, DEF_DIVIDERCHAR, DEF_BUSBITCHARS,
	DEF_PROPERTYDEFINITIONS, DEF_DEFAULTCAP,
	DEF_HISTORY, DEF_DIEAREA, DEF_COMPONENTS, DEF_VIAS,
	DEF_PINS, DEF_PINPROPERTIES, DEF_SPECIALNETS,
	DEF_NETS, DEF_IOTIMINGS, DEF_SCANCHAINS,
	DEF_CONSTRAINTS, DEF_GROUPS, DEF_EXTENSION,
	DEF_END};

void
DefRead(inName)
    char *inName;
{
    CellDef *rootDef;
    FILE *f;
    char *filename;
    char token[256];
    char *lineptr;
    int keyword, dscale, total;
    float oscale;

    static char *sections[] = {
	"VERSION",
	"NAMESCASESENSITIVE",
	"UNITS",
	"DESIGN",
	"REGIONS",
	"ROW",
	"TRACKS",
	"GCELLGRID",
	"DIVIDERCHAR",
	"BUSBITCHARS",
	"PROPERTYDEFINITIONS",
	"DEFAULTCAP",
	"HISTORY",
	"DIEAREA",
	"COMPONENTS",
	"VIAS",
	"PINS",
	"PINPROPERTIES",
	"SPECIALNETS",
	"NETS",
	"IOTIMINGS",
	"SCANCHAINS",
	"CONSTRAINTS",
	"GROUPS",
	"BEGINEXT",
	"END",
	NULL
    };

    f = lefFileOpen(NULL, inName, ".def", "r", &filename);

    if (f == NULL)
    {
#ifdef MAGIC_WRAPPER
	TxError("Cannot open input file %s (%s).\n", filename,
		strerror(errno));
#else
	TxError("Cannot open input file: ");
	perror(filename);
#endif
	return;
    }

    /* Initialize */

    TxPrintf("Reading DEF data from file %s.\n", filename);
    TxPrintf("This action is undoable.\n");
    UndoDisable();

    /* This works for CIF reads;  maybe should only do this if the top	*/
    /* cell is (UNNAMED)?						*/

    rootDef = EditCellUse->cu_def;
    DBCellRenameDef(rootDef, inName);
    oscale = CIFGetOutputScale(1000);
    lefCurrentLine = 0;

    /* Read file contents */

    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;	/* Blank line or null statement */

	keyword = Lookup(token, sections);
	if (keyword < 0)
	{
	    LefError("Unknown keyword \"%s\" in DEF file; ignoring.\n", token);
	    continue;
	}
	switch (keyword)
	{
	    case DEF_VERSION:
		break;
	    case DEF_NAMESCASESENSITIVE:
		break;
	    case DEF_REGIONS:
		LefSkipSection(f, sections[DEF_REGIONS]);
		break;
	    case DEF_DESIGN:
		break;
	    case DEF_UNITS:
		if (sscanf(lineptr, "%*s %*s %*s %d", &dscale) != 1)
		{
		    LefError("Invalid syntax for UNITS statement.\n");
		    LefError("Assuming default value of 100\n");
		    dscale = 100;
		}
		/* We don't care if the scale is 100, 200, 1000, or 2000. */
		/* Do we need to deal with numeric roundoff issues?	  */
		oscale *= (float)dscale;
		break;
	    case DEF_ROW:
		break;
	    case DEF_TRACKS:
		break;
	    case DEF_GCELLGRID:
		break;
	    case DEF_DIVIDERCHAR:
		break;
	    case DEF_BUSBITCHARS:
		break;
	    case DEF_HISTORY:
		/* HISTORY statements end only with a ";" character,	*/
		lineptr = LefComplete(f);	/* read to semicolon */
		break;
	    case DEF_DIEAREA:
		break;
	    case DEF_PROPERTYDEFINITIONS:
		LefSkipSection(f, sections[DEF_PROPERTYDEFINITIONS]);
		break;
	    case DEF_DEFAULTCAP:
		LefSkipSection(f, sections[DEF_DEFAULTCAP]);
		break;
	    case DEF_COMPONENTS:
		if (sscanf(lineptr, "%*s %d", &total) != 1) total = 0;
		DefReadComponents(f, rootDef, sections[DEF_COMPONENTS], oscale, total);
		break;
	    case DEF_VIAS:
		if (sscanf(lineptr, "%*s %d", &total) != 1) total = 0;
		LefSkipSection(f, sections[DEF_VIAS]);
		break;
	    case DEF_PINS:
		if (sscanf(lineptr, "%*s %d", &total) != 1) total = 0;
		LefSkipSection(f, sections[DEF_PINS]);
		break;
	    case DEF_PINPROPERTIES:
		LefSkipSection(f, sections[DEF_PINPROPERTIES]);
		break;
	    case DEF_SPECIALNETS:
		if (sscanf(lineptr, "%*s %d", &total) != 1) total = 0;
		DefReadNets(f, rootDef, sections[DEF_SPECIALNETS], oscale, TRUE, total);
		break;
	    case DEF_NETS:
		if (sscanf(lineptr, "%*s %d", &total) != 1) total = 0;
		DefReadNets(f, rootDef, sections[DEF_NETS], oscale, FALSE, total);
		break;
	    case DEF_IOTIMINGS:
		LefSkipSection(f, sections[DEF_IOTIMINGS]);
		break;
	    case DEF_SCANCHAINS:
		LefSkipSection(f, sections[DEF_SCANCHAINS]);
		break;
	    case DEF_CONSTRAINTS:
		LefSkipSection(f, sections[DEF_CONSTRAINTS]);
		break;
	    case DEF_GROUPS:
		LefSkipSection(f, sections[DEF_GROUPS]);
		break;
	    case DEF_EXTENSION:
		LefSkipSection(f, sections[DEF_EXTENSION]);
		break;
	    case DEF_END:
		if (!LefParseEndStatement(lineptr, "DESIGN"))
		{
		    LefError("END statement out of context.\n");
		    keyword = -1;
		}
		break;
	}
	if (keyword == DEF_END) break;
    }
    TxPrintf("DEF read: Processed %d lines.\n", lefCurrentLine);
    LefError(NULL);	/* print statement of errors, if any, and reset */

    /* Cleanup */

    DBAdjustLabels(rootDef, &TiPlaneRect);
    DBReComputeBbox(rootDef);
    DBWAreaChanged(rootDef, &rootDef->cd_bbox, DBW_ALLWINDOWS,
		&DBAllButSpaceBits);
    DBCellSetModified(rootDef, TRUE);

    UndoDisable();
    LefFreeInputBuffer();
    if (f != NULL) fclose(f);
}
